/*
 * OffStroke.c - main application file for OffStroke
 *
 * Copyright (C) 2003 Eric Ding
 *
 * Based on McPhling by Mike McCollister <McPhling@MikeMcCollister.com>.
 *
 * This file is part of OffStroke.
 *
 * OffStroke is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2, or (at your
 * option) any later version.
 *
 * OffStroke is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OffStroke; see the file COPYING.  If not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <PalmOS.h>

#include "OffStroke.h"
#include "OffStroke.rch"

/////////////
// Defines //
/////////////

#define ByteSwap16(n) ( ((((unsigned int) n) << 8) & 0xFF00) | \
                        ((((unsigned int) n) >> 8) & 0x00FF) )
#define TapSpotExtent 5

/////////////////////////
// Function Prototypes //
/////////////////////////

// Common Palm Functions
UInt16 StartApplication();
void EventLoop();
Boolean AppHandleEvent(EventPtr event);

// Utility Functions
UInt16 CheckOsVersion(Boolean bDisplayAlert);
UInt16 GetROMVer(UInt16 *ROMVerMajor, UInt16 *ROMVerMinor);
Boolean GetCollapsibleGraffiti();
void *Id2Ptr(UInt16 wID);
void LstSetBestWidth(UInt16 listID);

// Other Forward Declarations
Boolean MainEventHandler(EventPtr event);
Err MySysNotifyEventDequeuedEvent(SysNotifyParamType *np);
void EnableOffStroke(Boolean bEnabled);
Boolean HandleStroke(PointType* start, PointType* end);
void LoadAppPrefs(OffStrokePrefs *pPrefs);
void UpdateAppPrefs();

///////////////////
// Main Function //
///////////////////

UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP,
		 UInt16 launchFlags __attribute__ ((unused)))
{
    UInt16 error;
    OffStrokePrefs prefs;
    UInt16 myCardNo;
    LocalID myDbID;
	
    if (cmd == sysAppLaunchCmdNotify) {
	SysNotifyParamType *np = (SysNotifyParamType *)cmdPBP;

	switch(np->notifyType) {
	case sysNotifyEventDequeuedEvent:
	    // this is here in preparation for Palm OS 6 which will not allow
	    // the use of notifications with function pointers
	    MySysNotifyEventDequeuedEvent(np);
	    break;

	case sysNotifySyncStartEvent:
	    // turn off all notifications (makes HotSync faster)
	    EnableOffStroke(false);

            // enable sysNotifySyncFinishEvent (to turn back on notifications)
            SysCurAppDatabase(&myCardNo, &myDbID);
	    SysNotifyRegister(myCardNo, myDbID, sysNotifySyncFinishEvent,
			      NULL, sysNotifyNormalPriority, NULL);
	    break;

	case sysNotifySyncFinishEvent:
	    // turn on all notifications
	    EnableOffStroke(true);
	    // disable sysNotifySyncFinishEvent (to turn back on notifications)
	    SysCurAppDatabase(&myCardNo, &myDbID);
	    SysNotifyUnregister(myCardNo, myDbID, sysNotifySyncFinishEvent,
				NULL);
	    break;

	case sysNotifyDeleteProtectedEvent:
	{
	    SysNotifyDBInfoType *dbi =
		(SysNotifyDBInfoType *)np->notifyDetailsP;
			
	    // if "OffStroke" is being deleted, then unlock it
	    if (dbi->creator == APP_ID) {
		EnableOffStroke(false);
		np->handled = true;	
	    }
	}
	break;

	default:
	    break;
	}
    }
    else if (cmd == sysAppLaunchCmdNormalLaunch) {
	error = CheckOsVersion(true);
	if (error) {
	    if (GetROMVer(NULL, NULL) == 1)
		AppLaunchWithCommand(sysFileCDefaultApp,
				     sysAppLaunchCmdNormalLaunch, NULL);
	    return error;
	}

	StartApplication();
	EventLoop();
    }
    else if (cmd == sysAppLaunchCmdSyncNotify) {
	if (CheckOsVersion(false) == 0) {
	    LoadAppPrefs(&prefs);
	    EnableOffStroke(prefs.bEnabled);
	}
    }
    else if (cmd == sysAppLaunchCmdSystemReset) {
	if (CheckOsVersion(false) == 0) {
	    LoadAppPrefs(&prefs);
            if (prefs.bEnabled) {
		EnableOffStroke(prefs.bEnabled);
            }
	}
    }

    return 0;
}

UInt16 StartApplication()
{
    FrmGotoForm(MainForm);

    return 0;
}

void EventLoop()
{
    short err;
    EventType event;

    do {
	// wait for event or timeout after one second
	EvtGetEvent(&event, evtWaitForever);

	// handle any system events
	if (SysHandleEvent(&event))
	    continue;

	// handle any menu events (before I get messages)
	if (MenuHandleEvent(0, &event, &err))
	    continue;

	// handle application events
	if (AppHandleEvent(&event))
	    continue;

	// dispatch event to appropriate form
	FrmDispatchEvent(&event);

    } while (event.eType != appStopEvent);

    FrmCloseAllForms();
}

Boolean AppHandleEvent(EventPtr event)
{
    Int32 formID;
    FormPtr pForm;

    // load appropriate form
    if (event->eType == frmLoadEvent) {
	formID = event->data.frmLoad.formID;
	pForm = FrmInitForm(formID);
	FrmSetActiveForm(pForm);
	switch(formID) {
	case MainForm:
            FrmSetEventHandler(pForm, MainEventHandler);
            break;
	default:
            ErrNonFatalDisplayIf(true, "Invalid Form Load Event");
            return true;
            break;
	}
    }

    return false;
}

void ShowStrokePopups(FormPtr pForm)
{
    FrmShowObject(pForm, FrmGetObjectIndex(pForm, LabelStrokeFrom));
    FrmShowObject(pForm, FrmGetObjectIndex(pForm,
					   PopupTriggerStrokeFrom));

    FrmShowObject(pForm, FrmGetObjectIndex(pForm, LabelStrokeTo));
    FrmShowObject(pForm, FrmGetObjectIndex(pForm,
					   PopupTriggerStrokeTo));
}

void HideStrokePopups(FormPtr pForm)
{
    FrmHideObject(pForm, FrmGetObjectIndex(pForm, LabelStrokeFrom));
    FrmHideObject(pForm, FrmGetObjectIndex(pForm,
					   PopupTriggerStrokeFrom));

    FrmHideObject(pForm, FrmGetObjectIndex(pForm, LabelStrokeTo));
    FrmHideObject(pForm, FrmGetObjectIndex(pForm,
					   PopupTriggerStrokeTo));
}

void ShowTapPopups(FormPtr pForm)
{
    FrmShowObject(pForm, FrmGetObjectIndex(pForm, LabelTapSpot));
    FrmShowObject(pForm, FrmGetObjectIndex(pForm,
					   PopupTriggerTapSpot));
}

void HideTapPopups(FormPtr pForm)
{
    FrmHideObject(pForm, FrmGetObjectIndex(pForm, LabelTapSpot));
    FrmHideObject(pForm, FrmGetObjectIndex(pForm,
					   PopupTriggerTapSpot));
}

void UpdateForm(FormPtr pForm, OffStrokePrefs *pPrefs)
{
    ListPtr pList;
 
    CtlSetValue(Id2Ptr(CheckboxEnabled), pPrefs->bEnabled);

    CtlSetValue(Id2Ptr(CheckboxStroke), pPrefs->bStroke);

    pList = Id2Ptr(ListStrokeFrom);
    LstSetSelection(pList, pPrefs->strokeFrom);
    CtlSetLabel(Id2Ptr(PopupTriggerStrokeFrom),
		LstGetSelectionText(pList, pPrefs->strokeFrom));

    pList = Id2Ptr(ListStrokeTo);
    LstSetSelection(pList, pPrefs->strokeTo);
    CtlSetLabel(Id2Ptr(PopupTriggerStrokeTo),
		LstGetSelectionText(pList, pPrefs->strokeTo));

    CtlSetValue(Id2Ptr(CheckboxTap), pPrefs->bTap);
    FrmShowObject(pForm, FrmGetObjectIndex(pForm, CheckboxTap));

    pList = Id2Ptr(ListTapSpot);
    LstSetSelection(pList, pPrefs->tapSpot);
    CtlSetLabel(Id2Ptr(PopupTriggerTapSpot),
		LstGetSelectionText(pList, pPrefs->tapSpot));

    if (pPrefs->bEnabled) {
	FrmShowObject(pForm, FrmGetObjectIndex(pForm, CheckboxStroke));
	if (pPrefs->bStroke)
	    ShowStrokePopups(pForm);
	else
	    HideStrokePopups(pForm);

	if (pPrefs->bTap)
	    ShowTapPopups(pForm);
	else
	    HideTapPopups(pForm);
    } 
    else {
	FrmHideObject(pForm, FrmGetObjectIndex(pForm, CheckboxStroke));
	HideStrokePopups(pForm);
	FrmHideObject(pForm, FrmGetObjectIndex(pForm, CheckboxTap));
	HideTapPopups(pForm);
    }
}

void MainFormInit(FormPtr pForm)
{
    OffStrokePrefs *pPrefs;

    pPrefs = MemPtrNew(sizeof(OffStrokePrefs));
    LoadAppPrefs(pPrefs);
    FtrSet(APP_ID, FTR_PREFS, (UInt32)pPrefs);

    // Set list widths
    LstSetBestWidth(ListStrokeFrom);
    LstSetBestWidth(ListStrokeTo);

    // update form
    UpdateForm(pForm, pPrefs);
}

Boolean MainEventHandler(EventPtr event)
{
    Boolean handled = true;
    FormPtr pForm = FrmGetActiveForm();
    OffStrokePrefs *pPrefs;

    switch(event->eType) {
    case frmOpenEvent:
	// initialize and draw the form
	MainFormInit(pForm);
	FrmDrawForm(pForm);
	break;

#if 0
    // I don't think I really need this, but I keep it here and ifdef'ed out
    // just in case
    case displayExtentChangedEvent:
    case winDisplayChangedEvent:
	SizeForm(pForm);
	break;
#endif

    case frmUpdateEvent:
	FtrGet(APP_ID, FTR_PREFS, (UInt32 *)&pPrefs);
	UpdateForm(pForm, pPrefs);
	FrmDrawForm(pForm);
	break;

    case ctlSelectEvent:
	switch(event->data.ctlEnter.controlID) {
	case CheckboxEnabled:
	case CheckboxStroke:
	case CheckboxTap:
	    EnableOffStroke(CtlGetValue(Id2Ptr(CheckboxEnabled)));
	    UpdateAppPrefs();
	    FtrGet(APP_ID, FTR_PREFS, (UInt32 *)&pPrefs);
	    UpdateForm(pForm, pPrefs);
	    break;
	default:
	    handled = false;
	    break;
	}
	break;

    case menuEvent:
	switch(event->data.menu.itemID) {
	case menuitemID_about:
	{
	    FormPtr pForm = FrmGetActiveForm();
	    FormPtr pNewForm = FrmInitForm(AboutForm);

	    FrmDoDialog(pNewForm);
	    FrmDeleteForm(pNewForm);
	    FrmSetActiveForm(pForm);
	    break;
	}
	    
	default:
	    handled = false;
	    break;
	}
	break;

    case popSelectEvent:
	UpdateAppPrefs();
	handled = false;
	break;

    case frmCloseEvent:
	FrmEraseForm(pForm);
	FrmDeleteForm(pForm);

	if (FtrGet(APP_ID, FTR_PREFS, (UInt32 *)&pPrefs) == 0)
            if (pPrefs)
		MemPtrFree(pPrefs);
	FtrUnregister(APP_ID, FTR_PREFS);

	break;

    default:
	handled = false;
	break;
   }
   
   return handled;
}

Err MySysNotifyEventDequeuedEvent(SysNotifyParamType *np)
{
    EventPtr ep = (EventPtr)np->notifyDetailsP;

    if (ep->eType == ByteSwap16(penUpEvent)) {
	struct _PenUpEventType *pup = &(ep->data.penUp);
	PointType startPt, endPt;

	startPt.x = ByteSwap16(pup->start.x);
	startPt.y = ByteSwap16(pup->start.y);
	endPt.x   = ByteSwap16(pup->end.x);
	endPt.y   = ByteSwap16(pup->end.y);
 
	np->handled = HandleStroke(&startPt, &endPt);
	if (np->handled) {
	    MemSet(ep, sizeof(EventType), 0);
	    ep->eType = ByteSwap16(nilEvent);
	}
    }
    else {
	np->handled = false;
    }

    return 0;
}

void EnableOffStroke(Boolean bEnabled)
{
    UInt16 cardNo;
    LocalID dbID;
    MemHandle hCode = DmGetResource(sysResTAppCode, 0x0001);
    UInt32 temp;
    Boolean bNotificationsOn = false;

    if (FtrGet(APP_ID, FTR_NOTIFICATIONSON, &temp) != ftrErrNoSuchFeature)
	bNotificationsOn = true;

    SysCurAppDatabase(&cardNo, &dbID);

    if (bEnabled) {
	if (bNotificationsOn == false) {
	    UInt16 numButtons, numAreas;
	    const PenBtnInfoType *pPenBtnInfo = EvtGetPenBtnList(&numButtons);
	    const SilkscreenAreaType *pSilkscreenArea =
		EvtGetSilkscreenAreaList(&numAreas);

	    FtrSet(APP_ID, FTR_NOTIFICATIONSON, temp);

	    // lock myself
	    if (hCode) {
		MemHandleLock(hCode);
		DmDatabaseProtect(cardNo, dbID, true);
		DmReleaseResource(hCode);
	    }

	    // enable notifications
	    SysNotifyRegister(cardNo, dbID, sysNotifyEventDequeuedEvent,
			      &MySysNotifyEventDequeuedEvent,
			      sysNotifyNormalPriority, NULL);
	    SysNotifyRegister(cardNo, dbID, sysNotifySyncStartEvent,
			      NULL, sysNotifyNormalPriority, NULL);
	    SysNotifyRegister(cardNo, dbID, sysNotifyDeleteProtectedEvent,
			      NULL, sysNotifyNormalPriority, NULL);

	    // save button and silk locations
	    FtrSet(APP_ID, FTR_NUMBUTTONS, numButtons);
	    FtrPtrNew(APP_ID, FTR_BUTTONS, numButtons * sizeof(PenBtnInfoType),
		      (void **)&pPenBtnInfo);
	    FtrSet(APP_ID, FTR_NUMSILK, numAreas);
	    FtrPtrNew(APP_ID, FTR_SILK, numAreas * sizeof(SilkscreenAreaType),
		      (void **)&pSilkscreenArea);

	    // save collapsible graffiti state
	    if (GetCollapsibleGraffiti())
		FtrSet(APP_ID, FTR_PSCOLLGRAFFITI, NULL);
	    else
		FtrUnregister(APP_ID, FTR_PSCOLLGRAFFITI);
	}
    }
    else {
	if (bNotificationsOn) {
	    FtrUnregister(APP_ID, FTR_NOTIFICATIONSON);

	    // unlock myself
	    if (hCode) {
		DmDatabaseProtect(cardNo, dbID, false);
		MemHandleUnlock(hCode);
		DmReleaseResource(hCode);
	    }
	}

	// still disable notifications just in case
	SysNotifyUnregister(cardNo, dbID, sysNotifyEventDequeuedEvent,
			    NULL);
	SysNotifyUnregister(cardNo, dbID, sysNotifySyncStartEvent,
			    NULL);
	SysNotifyUnregister(cardNo, dbID, sysNotifyDeleteProtectedEvent,
			    NULL);

	// free button and silk locations
	if (FtrGet(APP_ID, FTR_NUMBUTTONS, &temp) == 0)
	    FtrUnregister(APP_ID, FTR_NUMBUTTONS);
	if (FtrGet(APP_ID, FTR_BUTTONS, &temp) == 0)
	    FtrPtrFree(APP_ID, FTR_BUTTONS);
	if (FtrGet(APP_ID, FTR_NUMSILK, &temp) == 0)
	    FtrUnregister(APP_ID, FTR_NUMSILK);
	if (FtrGet(APP_ID, FTR_SILK, &temp) == 0)
	    FtrPtrFree(APP_ID, FTR_SILK);
    }
}

void GetStrokeBound(UInt16 location, RectangleType *pBound,
		    RectangleType *pRectApps,
		    RectangleType *pRectMenu,
		    RectangleType *pRectCalc,
		    RectangleType *pRectFind,
		    RectangleType *pRectGraffiti,
		    RectangleType *pRectScreen)
{
    switch((StrokeToType)location) {
    case sttApps:
	*pBound = *pRectApps;
	break;

    case sttMenu:
	*pBound = *pRectMenu;
	break;

    case sttCalc:
	*pBound = *pRectCalc;
	break;

    case sttFind:
	*pBound = *pRectFind;
	break;

    case sttGraffiti:
	*pBound = *pRectGraffiti;
	break;

    case sttScreen:
	*pBound = *pRectScreen;
	break;
    }
}

void GetStrokeBounds(OffStrokePrefs *pPrefs,
		     RectangleType *pStrokeFrom,
		     RectangleType *pStrokeTo,
		     RectangleType *pRectAppsButton)
{
    RectangleType rectApps            = {{  0, 161}, { 24,  32}};
    RectangleType rectMenu            = {{  0, 191}, { 24,  32}};
    RectangleType rectCalc            = {{136, 161}, { 24,  32}};
    RectangleType rectFind            = {{136, 191}, { 24,  32}};
    RectangleType rectGraffiti        = {{ 27, 164}, {106,  56}};
    RectangleType rectGraffitiAlpha   = {{ 27, 164}, { 62,  56}};
    RectangleType rectGraffitiNumeric = {{ 89, 164}, { 44,  56}};
    RectangleType rectScreen          = {{  0,   0}, {MAX_WIDTH, MAX_HEIGHT}};

    UInt16 i;
    UInt16 numButtons = 0, numAreas = 0;
    PenBtnInfoType *penBtnInfo;
    SilkscreenAreaType *silkscreenArea;
    UInt32 temp;

    // get button and silk areas
    if(FtrGet(APP_ID, FTR_PSCOLLGRAFFITI, &temp) == ftrErrNoSuchFeature) {
	if(FtrGet(APP_ID, FTR_NUMBUTTONS, &temp) == 0) {
	    numButtons = temp;
	    FtrGet(APP_ID, FTR_BUTTONS, (UInt32 *)&penBtnInfo);
	}
	if(FtrGet(APP_ID, FTR_NUMSILK, &temp) == 0) {
	    numAreas = temp;
	    FtrGet(APP_ID, FTR_SILK, (UInt32 *)&silkscreenArea);
	}
    }
    else {
	penBtnInfo = (PenBtnInfoType *)EvtGetPenBtnList(&numButtons);
	silkscreenArea =
	    (SilkscreenAreaType *)EvtGetSilkscreenAreaList(&numAreas);
    }

    // assign rect values for buttons
    for(i = 0; i < numButtons ; i++) {
	// need to take into account multiple copies of rectApps, rectMenu
	// rectCalc and rectFind
         
	switch(penBtnInfo[i].asciiCode) {
	case vchrLaunch:
	    rectApps = penBtnInfo[i].boundsR;
	    break;
	case vchrMenu:
	    rectMenu = penBtnInfo[i].boundsR;
	    break;
	case vchrCalc:
	    rectCalc = penBtnInfo[i].boundsR;
	    break;
	case vchrFind:
	    rectFind = penBtnInfo[i].boundsR;
	    break;
	}
    }

    // assign rect values for silkscreen
    for(i = 0; i < numAreas ; i++) {
	switch(silkscreenArea[i].areaType) {
	case silkscreenRectGraffiti:
	    switch(silkscreenArea[i].index) {
	    case alphaGraffitiSilkscreenArea:
		rectGraffitiAlpha = silkscreenArea[i].bounds;
		break;
	       
	    case numericGraffitiSilkscreenArea:
		rectGraffitiNumeric = silkscreenArea[i].bounds;
		break;
	    }
	    break;
	   
	case silkscreenRectScreen:
	    rectScreen = silkscreenArea[i].bounds;
	    break;
	}
    }

    // set whole graffiti rectangle
    if(rectGraffitiAlpha.topLeft.y < rectGraffitiNumeric.topLeft.y) {
	rectGraffiti           = rectGraffitiAlpha;
	rectGraffiti.extent.y += rectGraffitiNumeric.extent.y;
    }
    else if(rectGraffitiAlpha.topLeft.y > rectGraffitiNumeric.topLeft.y) {
	rectGraffiti           = rectGraffitiNumeric;
	rectGraffiti.extent.y += rectGraffitiAlpha.extent.y;
    }
    else {
	rectGraffiti           = rectGraffitiAlpha;
	rectGraffiti.extent.x += rectGraffitiNumeric.extent.x;
    }

    // get bounds
    GetStrokeBound(pPrefs->strokeFrom, pStrokeFrom,
		   &rectApps, &rectMenu, &rectCalc, &rectFind,
		   &rectGraffiti, &rectScreen);
    GetStrokeBound(pPrefs->strokeTo, pStrokeTo,
		   &rectApps, &rectMenu, &rectCalc, &rectFind,
		   &rectGraffiti, &rectScreen);

    *pRectAppsButton = rectApps;
}

void TurnOffPalm()
{
    if (PrefGetPreference(prefAutoLockType) == uponPowerOff) {
	EvtEnqueueKey(vchrLock, 0, commandKeyMask);
    }
    else {
	EvtEnqueueKey(vchrPowerOff, 0, commandKeyMask);
    }
}

Boolean HandleStroke(PointType* start, PointType* end)
{
    UInt32 temp;
    OffStrokePrefs prefs;
    RectangleType rectAppsButton;

    // don't try if in the midst of updating preferences
    if (FtrGet(APP_ID, FTR_PREFSTEMP, (UInt32 *)&temp) != ftrErrNoSuchFeature)
	return false;

    LoadAppPrefs(&prefs);
    if (prefs.bEnabled && (prefs.bStroke || prefs.bTap)) {
	RectangleType strokeFrom, strokeTo;
	Boolean bGraffitiAreaVisible = true;
	Coord x, y;

	// adjust if graffiti area is not visible
	WinGetDisplayExtent(&x, &y);

	if (prefs.bTap) {
	    RectangleType tapRect;
	    tapRect.extent.x = tapRect.extent.y = TapSpotExtent;
	    switch (prefs.tapSpot) {
	    case topLeft:
		tapRect.topLeft.x = tapRect.topLeft.y = 0;
		break;
	    case topRight:
		tapRect.topLeft.x = x - TapSpotExtent;
		tapRect.topLeft.y = 0;
		break;
	    case bottomLeft:
		tapRect.topLeft.x = 0;
		tapRect.topLeft.y = y - TapSpotExtent;
		break;
	    case bottomRight:
		tapRect.topLeft.x = x - TapSpotExtent;
		tapRect.topLeft.y = y - TapSpotExtent;
		break;
	    }
	    if (RctPtInRectangle(start->x, start->y, &tapRect) &&
		RctPtInRectangle(end->x, end->y, &tapRect)) {
		TurnOffPalm();
		return true;
	    }
	}

	if (prefs.bStroke) {
	    GetStrokeBounds(&prefs, &strokeFrom, &strokeTo, &rectAppsButton);

	    if ((rectAppsButton.topLeft.y < y))
		bGraffitiAreaVisible = false;

	    if (bGraffitiAreaVisible) {
		if (RctPtInRectangle(start->x, start->y, &strokeFrom) &&
		    RctPtInRectangle(end->x, end->y, &strokeTo)) {
		    TurnOffPalm();
		    return true;
		}
	    }
	}
    }
    return false;
}

void GetCurrentPrefs(OffStrokePrefs *pPrefs)
{
    // get the preferences from the screen
    pPrefs->bEnabled = CtlGetValue(Id2Ptr(CheckboxEnabled));

    if (pPrefs->bEnabled) {
	pPrefs->bStroke = CtlGetValue(Id2Ptr(CheckboxStroke));
	if (pPrefs->bStroke) {
	    pPrefs->strokeFrom = LstGetSelection(Id2Ptr(ListStrokeFrom));
	    pPrefs->strokeTo = LstGetSelection(Id2Ptr(ListStrokeTo));
	}

	pPrefs->bTap = CtlGetValue(Id2Ptr(CheckboxTap));
	if (pPrefs->bTap) {
	    pPrefs->tapSpot = LstGetSelection(Id2Ptr(ListTapSpot));
	}
    }
}

void UpdateAppPrefs()
{
    OffStrokePrefs *pPrefs, *pPrefsTemp;

    // get prefs pointer
    FtrGet(APP_ID, FTR_PREFS, (UInt32 *)&pPrefs);

    // set up prefs temp pointer
    pPrefsTemp = MemPtrNew(sizeof(OffStrokePrefs));
    *pPrefsTemp = *pPrefs;
    FtrSet(APP_ID, FTR_PREFSTEMP, (UInt32)pPrefsTemp);

    // get data from form
    GetCurrentPrefs(pPrefsTemp);

    // copy data to final destination
    pPrefs->bEnabled = pPrefsTemp->bEnabled;
    pPrefs->bStroke = pPrefsTemp->bStroke;
    pPrefs->strokeFrom = pPrefsTemp->strokeFrom;
    pPrefs->strokeTo = pPrefsTemp->strokeTo;
    pPrefs->bTap = pPrefsTemp->bTap;
    pPrefs->tapSpot = pPrefsTemp->tapSpot;

    // save the prefs data
    PrefSetAppPreferences(APP_ID, PREFS, 0, pPrefs, sizeof(OffStrokePrefs),
			  true);

    MemPtrFree(pPrefsTemp);
    FtrUnregister(APP_ID, FTR_PREFSTEMP);
}

void LoadAppPrefs(OffStrokePrefs *pPrefs)
{
    UInt16 retSize = 0;
    Int16 prefResult;
    UInt16 prefsSize = sizeof(OffStrokePrefs);

    prefResult = PrefGetAppPreferences(APP_ID, PREFS, NULL,
				       &retSize, true);
    if ((prefResult == noPreferenceFound) || (retSize != prefsSize)) {
	pPrefs->prefVersion = 0x01;
	pPrefs->bEnabled = true;
	pPrefs->bStroke = true;
	pPrefs->strokeFrom = sftApps;
	pPrefs->strokeTo = sttMenu;
	pPrefs->bTap = false;
	pPrefs->tapSpot = bottomLeft;
	PrefSetAppPreferences(APP_ID, PREFS, 0, pPrefs, prefsSize, true);
    }
    else
	PrefGetAppPreferences(APP_ID, PREFS, pPrefs, &retSize, true);
}

///////////////////////
// Utility Functions //
///////////////////////

UInt16 CheckOsVersion(Boolean bDisplayAlert)
{
    UInt16 error = 0;

    // check if PalmOS 5.0 or greater
    if (GetROMVer(NULL, NULL) < 5) {
	if (bDisplayAlert)
	    FrmCustomAlert(VersionAlert, "5.0", NULL, NULL);
	error = sysErrRomIncompatible;
    }

    return error;
}

UInt16 GetROMVer(UInt16 *ROMVerMajor, UInt16 *ROMVerMinor)
{
   UInt32 dwROMVer;

   FtrGet(sysFtrCreator, sysFtrNumROMVersion, &dwROMVer);

   if(ROMVerMajor != NULL)
      *ROMVerMajor = (UInt16)sysGetROMVerMajor(dwROMVer);
   if(ROMVerMinor != NULL)
      *ROMVerMinor = (UInt16)sysGetROMVerMinor(dwROMVer);

   return (UInt16)sysGetROMVerMajor(dwROMVer);
}

Boolean GetCollapsibleGraffiti()
{
    UInt32 pinMgrVersion;
    Boolean bPS = false;
    Err error;

    // PalmSource detection
    error = FtrGet(pinCreator, pinFtrAPIVersion, &pinMgrVersion);
    if (error == errNone) {
	if (pinMgrVersion == pinAPIVersion1_0) {
	    // make sure that it works with the Gramin iQue 3600
	    bPS = true;
	}
	else if (pinMgrVersion >= pinAPIVersion1_1) {
	    UInt32 flags;

	    error = FtrGet(sysFtrCreator, sysFtrNumInputAreaFlags, &flags);
	    if ((error == errNone) && 
		(flags & grfFtrInputAreaFlagCollapsible)) {
		bPS = true;
	    }
	}
    }

    return bPS;
}

void *Id2Ptr(UInt16 wID)
{
   FormPtr pForm = FrmGetActiveForm();

   return(FrmGetObjectPtr(pForm, FrmGetObjectIndex(pForm, wID)));
}

void LstSetBestWidth(UInt16 listID)
{
    Int16 i;
    FormPtr pForm = FrmGetActiveForm();
    ListType *pList = Id2Ptr(listID);
    Int16 numItems = LstGetNumberOfItems(pList);
    UInt16 listIndex = FrmGetObjectIndex(pForm, listID);
    UInt16 width, maxWidth = 0;
    RectangleType listBounds;

    for(i = 0; i < numItems; i++) {
	width = StrLen(LstGetSelectionText(pList, i));
	width = FntCharsWidth(LstGetSelectionText(pList, i), width);
	if(width > maxWidth)
	    maxWidth = width;
    }

    // add width to account for up/down arrows
    if(numItems > LstGetVisibleItems(pList)) {
	Char downArrow = symbolSmallDownArrow;

	FntSetFont(symbolFont);
	maxWidth += FntCharsWidth(&downArrow, 1);
	maxWidth -= LIST_PADDING;
	FntSetFont(stdFont);
    }

    FrmGetObjectBounds(pForm, listIndex, &listBounds);
    listBounds.extent.x = maxWidth + LIST_PADDING;
    FrmSetObjectBounds(pForm, listIndex, &listBounds);
}
